/* SystemTap probe support for GDB.

   Copyright (C) 2011 Free Software Foundation, Inc.

   This file is part of GDB.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#include "defs.h"
#include "stap-probe.h"
#include "vec.h"
#include "ui-out.h"
#include "gdb_regex.h"
#include "objfiles.h"
#include "arch-utils.h"
#include "command.h"
#include "gdbcmd.h"
#include "filenames.h"
#include "value.h"
#include "exceptions.h"
#include "ax.h"
#include "ax-gdb.h"
#include "complaints.h"
#include "cli/cli-utils.h"
#include "linespec.h"
#include "user-regs.h"
#include "parser-defs.h"
#include "language.h"

#include <ctype.h>

/* The maximum number of arguments that a probe can have,
   as defined in <sys/sdt.h>.  */

#define STAP_MAX_ARGS 12

/* Should we display debug information for the probe's argument expression
   parsing?  */

static int stap_expression_debug = 0;

/* The various possibilities of bitness defined for a probe's argument.

   The relationship is:

   - STAP_ARG_BITNESS_UNDEFINED:  The user hasn't specified the bitness.
   - STAP_ARG_BITNESS_32BIT_UNSIGNED:  argument string starts with `4@'.
   - STAP_ARG_BITNESS_32BIT_SIGNED:  argument string starts with `-4@'.
   - STAP_ARG_BITNESS_64BIT_UNSIGNED:  argument string starts with `8@'.
   - STAP_ARG_BITNESS_64BIT_SIGNED:  argument string starts with `-8@'.  */

enum stap_arg_bitness
{
  STAP_ARG_BITNESS_UNDEFINED,
  STAP_ARG_BITNESS_32BIT_UNSIGNED,
  STAP_ARG_BITNESS_32BIT_SIGNED,
  STAP_ARG_BITNESS_64BIT_UNSIGNED,
  STAP_ARG_BITNESS_64BIT_SIGNED,
};

/* The following structure represents a single argument for the probe.  */

struct stap_probe_arg
{
  /* The bitness of this argument.  */
  enum stap_arg_bitness bitness;

  /* The corresponding `struct type *' to the bitness.  */
  struct type *atype;

  /* The argument converted to an internal GDB expression.  */
  struct expression *aexpr;
};

/* Structure that holds information about all arguments of a probe.  */

struct stap_args_info
{
  /* The number of valid parsed arguments.  */
  int n_args;

  /* The probe to which these arguments belong.  */
  struct stap_probe *probe;

  /* Information about each argument.  This is an array of `stap_probe_arg',
     with each entry representing one argument.  */
  struct stap_probe_arg *args;
};

/* When parsing the arguments, we have to establish different precedences
   for the various kinds of asm operators.  This enumeration represents those
   precedences.

   This logic behind this is available at
   <http://sourceware.org/binutils/docs/as/Infix-Ops.html#Infix-Ops>, or using
   the command "info '(as)Infix Ops'".  */

enum stap_operand_prec
{
  /* Lowest precedence, used for non-recognized operands or for the beginning
     of the parsing process.  */
  STAP_OPERAND_PREC_NONE = 0,

  /* Precedence of logical OR.  */
  STAP_OPERAND_PREC_LOGICAL_OR,

  /* Precedence of logical AND.  */
  STAP_OPERAND_PREC_LOGICAL_AND,

  /* Precedence of additive (plus, minus) and comparative (equal, less,
     greater-than, etc) operands.  */
  STAP_OPERAND_PREC_ADD_CMP,

  /* Precedence of bitwise operands (bitwise OR, XOR, bitwise AND,
     logical NOT).  */
  STAP_OPERAND_PREC_BITWISE,

  /* Precedence of multiplicative operands (multiplication, division,
     remainder, left shift and right shift).  */
  STAP_OPERAND_PREC_MUL
};

/* This dummy variable is used when parsing a probe's argument fails.
   In this case, the number of arguments for this probe is zero, so that's
   why this variable is useful.  */

static struct stap_args_info dummy_stap_args_info =
  { 0, NULL, NULL };

static void stap_parse_argument_1 (struct stap_parse_info *p, int has_lhs,
				   enum stap_operand_prec prec);

static void stap_parse_argument_conditionally (struct stap_parse_info *p);

/* Returns 1 if *S is an operator, zero otherwise.  */

static int stap_is_operator (char op);

static void
show_stapexpressiondebug (struct ui_file *file, int from_tty,
			  struct cmd_list_element *c, const char *value)
{
  fprintf_filtered (file, _("SystemTap Probe expression debugging is %s.\n"),
		    value);
}

/* Returns the operator precedence level of OP, or STAP_OPERAND_PREC_NONE
   if the operator code was not recognized.  */

static enum stap_operand_prec
stap_get_operator_prec (enum exp_opcode op)
{
  switch (op)
    {
    case BINOP_LOGICAL_OR:
      return STAP_OPERAND_PREC_LOGICAL_OR;

    case BINOP_LOGICAL_AND:
      return STAP_OPERAND_PREC_LOGICAL_AND;

    case BINOP_ADD:
    case BINOP_SUB:
    case BINOP_EQUAL:
    case BINOP_NOTEQUAL:
    case BINOP_LESS:
    case BINOP_LEQ:
    case BINOP_GTR:
    case BINOP_GEQ:
      return STAP_OPERAND_PREC_ADD_CMP;

    case BINOP_BITWISE_IOR:
    case BINOP_BITWISE_AND:
    case BINOP_BITWISE_XOR:
    case UNOP_LOGICAL_NOT:
      return STAP_OPERAND_PREC_BITWISE;

    case BINOP_MUL:
    case BINOP_DIV:
    case BINOP_REM:
    case BINOP_LSH:
    case BINOP_RSH:
      return STAP_OPERAND_PREC_MUL;

    default:
      return STAP_OPERAND_PREC_NONE;
    }
}

/* Given S, read the operator in it and fills the OP pointer with its code.
   Return 1 on success, zero if the operator was not recognized.  */

static int
stap_get_opcode (const char **s, enum exp_opcode *op)
{
  const char c = **s;
  int ret = 1;

  *s += 1;

  switch (c)
    {
    case '*':
      *op = BINOP_MUL;
      break;

    case '/':
      *op = BINOP_DIV;
      break;

    case '%':
      *op = BINOP_REM;
    break;

    case '<':
      *op = BINOP_LESS;
      if (**s == '<')
	{
	  *s += 1;
	  *op = BINOP_LSH;
	}
      else if (**s == '=')
	{
	  *s += 1;
	  *op = BINOP_LEQ;
	}
      else if (**s == '>')
	{
	  *s += 1;
	  *op = BINOP_NOTEQUAL;
	}
    break;

    case '>':
      *op = BINOP_GTR;
      if (**s == '>')
	{
	  *s += 1;
	  *op = BINOP_RSH;
	}
      else if (**s == '=')
	{
	  *s += 1;
	  *op = BINOP_GEQ;
	}
    break;

    case '|':
      *op = BINOP_BITWISE_IOR;
      if (**s == '|')
	{
	  *s += 1;
	  *op = BINOP_LOGICAL_OR;
	}
    break;

    case '&':
      *op = BINOP_BITWISE_AND;
      if (**s == '&')
	{
	  *s += 1;
	  *op = BINOP_LOGICAL_AND;
	}
    break;

    case '^':
      *op = BINOP_BITWISE_XOR;
      break;

    case '!':
      *op = UNOP_LOGICAL_NOT;
      break;

    case '+':
      *op = BINOP_ADD;
      break;

    case '-':
      *op = BINOP_SUB;
      break;

    case '=':
      if (**s != '=')
	{
	  ret = 0;
	  break;
	}
      *op = BINOP_EQUAL;
      break;

    default:
      /* We didn't find any operator.  */
      *s -= 1;
      return 0;
    }

  return ret;
}

/* Given the bitness of the argument, represented by B, return the
   corresponding `struct type *'.  */

static struct type *
stap_get_expected_argument_type (struct gdbarch *gdbarch,
				 enum stap_arg_bitness b)
{
  switch (b)
    {
    case STAP_ARG_BITNESS_UNDEFINED:
      if (gdbarch_addr_bit (gdbarch) == 32)
	return builtin_type (gdbarch)->builtin_uint32;
      else
	return builtin_type (gdbarch)->builtin_uint64;

    case STAP_ARG_BITNESS_32BIT_SIGNED:
      return builtin_type (gdbarch)->builtin_int32;

    case STAP_ARG_BITNESS_32BIT_UNSIGNED:
      return builtin_type (gdbarch)->builtin_uint32;

    case STAP_ARG_BITNESS_64BIT_SIGNED:
      return builtin_type (gdbarch)->builtin_int64;

    case STAP_ARG_BITNESS_64BIT_UNSIGNED:
      return builtin_type (gdbarch)->builtin_uint64;

    default:
      internal_error (__FILE__, __LINE__,
		      _("Undefined bitness for probe."));
      break;
    }
}

static void
stap_parse_register_operand (struct stap_parse_info *p)
{
  /* Simple flag to indicate whether we have seen a minus signal before
     certain number.  */
  int got_minus = 0;
  /* Flags to indicate whether this register access is being displaced and/or
     indirected.  */
  int disp_p = 0, indirect_p = 0;
  struct gdbarch *gdbarch = p->gdbarch;
  /* Needed to generate the register name as a part of an expression.  */
  struct stoken str;
  /* Variables used to extract the register name from the probe's
     argument.  */
  const char *start;
  char *regname;
  int len;

  /* Prefixes for the parser.  */
  const char *reg_prefix = gdbarch_stap_register_prefix (gdbarch);
  const char *reg_ind_prefix
    = gdbarch_stap_register_indirection_prefix (gdbarch);
  const char *gdb_reg_prefix = gdbarch_stap_gdb_register_prefix (gdbarch);
  int reg_prefix_len = reg_prefix ? strlen (reg_prefix) : 0;
  int reg_ind_prefix_len = reg_ind_prefix ? strlen (reg_ind_prefix) : 0;
  int gdb_reg_prefix_len = gdb_reg_prefix ? strlen (gdb_reg_prefix) : 0;

  /* Sufixes for the parser.  */
  const char *reg_sufix = gdbarch_stap_register_sufix (gdbarch);
  const char *reg_ind_sufix
    = gdbarch_stap_register_indirection_sufix (gdbarch);
  const char *gdb_reg_sufix = gdbarch_stap_gdb_register_sufix (gdbarch);
  int reg_sufix_len = reg_sufix ? strlen (reg_sufix) : 0;
  int reg_ind_sufix_len = reg_ind_sufix ? strlen (reg_ind_sufix) : 0;
  int gdb_reg_sufix_len = gdb_reg_sufix ? strlen (gdb_reg_sufix) : 0;

  /* Checking for a displacement argument.  */
  if (*p->arg == '+')
    /* If it's a plus sign, we don't need to do anything, just advance the
       pointer.  */
    ++p->arg;

  if (*p->arg == '-')
    {
      got_minus = 1;
      ++p->arg;
    }

  if (isdigit (*p->arg))
    {
      /* The value of the displacement.  */
      long displacement;

      disp_p = 1;
      displacement = strtol (p->arg, (char **) &p->arg, 10);

      /* Generating the expression for the displacement.  */
      write_exp_elt_opcode (OP_LONG);
      write_exp_elt_type (builtin_type (gdbarch)->builtin_long);
      write_exp_elt_longcst (displacement);
      write_exp_elt_opcode (OP_LONG);
      if (got_minus)
	write_exp_elt_opcode (UNOP_NEG);
    }

  /* Getting rid of register indirection prefix.  */
  if (reg_ind_prefix
      && strncmp (p->arg, reg_ind_prefix, reg_ind_prefix_len) == 0)
    {
      indirect_p = 1;
      p->arg += reg_ind_prefix_len;
    }

  if (disp_p && !indirect_p)
    error (_("Invalid register displacement syntax on expression `%s'."),
	   p->saved_arg);

  /* Getting rid of register prefix.  */
  if (reg_prefix && strncmp (p->arg, reg_prefix, reg_prefix_len) == 0)
    p->arg += reg_prefix_len;

  /* Now we should have only the register name.  Let's extract it and get
     the associated number.  */
  start = p->arg;

  /* We assume the register name is composed by letters and numbers.  */
  while (isalnum (*p->arg))
    ++p->arg;

  len = p->arg - start;

  regname = alloca (len + gdb_reg_prefix_len + gdb_reg_sufix_len + 1);
  regname[0] = '\0';

  /* We only add the GDB's register prefix/sufix if we are dealing with
     a numeric register.  */
  if (gdb_reg_prefix && isdigit (*start))
    {
      strncpy (regname, gdb_reg_prefix, gdb_reg_prefix_len);
      strncpy (regname + gdb_reg_prefix_len, start, len);

      if (gdb_reg_sufix)
	strncpy (regname + gdb_reg_prefix_len + len,
		 gdb_reg_sufix, gdb_reg_sufix_len);

      len += gdb_reg_prefix_len + gdb_reg_sufix_len;
    }
  else
    strncpy (regname, start, len);

  regname[len] = '\0';
  /* Is this a valid register name?  */
  if (user_reg_map_name_to_regnum (gdbarch, regname, len) == -1)
    error (_("Invalid register name `%s' on expression `%s'."),
	   regname, p->saved_arg);

  write_exp_elt_opcode (OP_REGISTER);
  str.ptr = regname;
  str.length = len;
  write_exp_string (str);
  write_exp_elt_opcode (OP_REGISTER);

  if (indirect_p)
    {
      if (disp_p)
	write_exp_elt_opcode (BINOP_ADD);

      /* Casting to the expected type.  */
      write_exp_elt_opcode (UNOP_CAST);
      write_exp_elt_type (lookup_pointer_type (p->arg_type));
      write_exp_elt_opcode (UNOP_CAST);

      write_exp_elt_opcode (UNOP_IND);
    }

  /* Getting rid of the register name sufix.  */
  if (reg_sufix)
    {
      if (strncmp (p->arg, reg_sufix, reg_sufix_len) != 0)
	error (_("Missing register name sufix `%s' on expression `%s'."),
	       reg_sufix, p->saved_arg);

      p->arg += reg_sufix_len;
    }

  /* Getting rid of the register indirection sufix.  */
  if (indirect_p && reg_ind_sufix)
    {
      if (strncmp (p->arg, reg_ind_sufix, reg_ind_sufix_len) != 0)
	error (_("Missing indirection sufix `%s' on expression `%s'."),
	       reg_ind_sufix, p->saved_arg);

      p->arg += reg_ind_sufix_len;
    }
}

static void
stap_parse_single_operand (struct stap_parse_info *p)
{
  struct gdbarch *gdbarch = p->gdbarch;
  /* Prefixes for the parser.  */
  const char *const_prefix = gdbarch_stap_integer_prefix (gdbarch);
  const char *reg_prefix = gdbarch_stap_register_prefix (gdbarch);
  const char *reg_ind_prefix
    = gdbarch_stap_register_indirection_prefix (gdbarch);
  int const_prefix_len = const_prefix ? strlen (const_prefix) : 0;
  int reg_prefix_len = reg_prefix ? strlen (reg_prefix) : 0;
  int reg_ind_prefix_len = reg_ind_prefix ? strlen (reg_ind_prefix) : 0;

  /* Sufixes for the parser.  */
  const char *const_sufix = gdbarch_stap_integer_sufix (gdbarch);
  const char *reg_sufix = gdbarch_stap_register_sufix (gdbarch);
  const char *reg_ind_sufix
    = gdbarch_stap_register_indirection_sufix (gdbarch);
  int const_sufix_len = const_sufix ? strlen (const_sufix) : 0;
  int reg_sufix_len = reg_sufix ? strlen (reg_sufix) : 0;
  int reg_ind_sufix_len = reg_ind_sufix ? strlen (reg_ind_sufix) : 0;

  /* We first try to parse this token as a "special token".  */
  if (gdbarch_stap_parse_special_token_p (gdbarch))
    {
      int ret = gdbarch_stap_parse_special_token (gdbarch, p);

      if (ret)
	/* If the return value of the above function is not zero,
	   it means it successfully parsed the special token.

	   If it is NULL, we try to parse it using our method.  */
	return;
    }

  if (*p->arg == '-' || *p->arg == '~' || *p->arg == '+')
    {
      char c = *p->arg;
      /* We use this variable to do a lookahead.  */
      const char *tmp = p->arg;

      ++tmp;

      /* This is an unary operation.  Here is a list of allowed tokens
	 here:

	 - numeric literal;
	 - number (from register displacement)
	 - subexpression (beginning with `(')

	 We handle the register displacement here, and the other cases
	 recursively.  */
      if (isdigit (*tmp))
	{
	  int number = strtol (tmp, (char **) &tmp, 10);

	  if (p->inside_paren_p)
	    tmp = skip_spaces_const (tmp);
	  if (!reg_ind_prefix
	      || strncmp (tmp, reg_ind_prefix, reg_ind_prefix_len) != 0)
	    goto not_displacement;

	  /* If we are here, it means it is a displacement.  The only
	     operations allowed here are `-' and `+'.  */
	  if (c == '~')
	    error (_("Invalid operator `%c' for register displacement "
		     "on expression `%s'."), c, p->saved_arg);

	  stap_parse_register_operand (p);
	}
      else
not_displacement:
	{
	  p->arg = tmp;
	  stap_parse_argument_conditionally (p);
	  if (c == '-')
	    write_exp_elt_opcode (UNOP_NEG);
	  else if (c == '~')
	    write_exp_elt_opcode (UNOP_COMPLEMENT);
	}
    }
  else if (isdigit (*p->arg))
    {
      /* A temporary variable, needed for lookahead.  */
      const char *tmp = p->arg;
      long number;

      /* We can be dealing with a numeric constant (if `const_prefix' is
	 NULL), or with a register displacement.  */
      number = strtol (tmp, (char **) &tmp, 10);

      if (p->inside_paren_p)
	tmp = skip_spaces_const (tmp);
      if (!const_prefix && reg_ind_prefix
	  && strncmp (tmp, reg_ind_prefix, reg_ind_prefix_len) != 0)
	{
	  /* We are dealing with a numeric constant.  */
	  write_exp_elt_opcode (OP_LONG);
	  write_exp_elt_type (builtin_type (gdbarch)->builtin_long);
	  write_exp_elt_longcst (number);
	  write_exp_elt_opcode (OP_LONG);

	  p->arg = tmp;

	  if (const_sufix)
	    {
	      if (strncmp (p->arg, const_sufix, const_sufix_len) == 0)
		p->arg += const_sufix_len;
	      else
		error (_("Invalid constant sufix on expression `%s'."),
		       p->saved_arg);
	    }
	}
      else if (reg_ind_prefix
	       && strncmp (tmp, reg_ind_prefix, reg_ind_prefix_len) == 0)
	stap_parse_register_operand (p);
      else
	error (_("Unknown numeric token on expression `%s'."),
	       p->saved_arg);
    }
  else if (const_prefix
	   && strncmp (p->arg, const_prefix, const_prefix_len) == 0)
    {
      /* We are dealing with a numeric constant.  */
      long number;

      p->arg += const_prefix_len;
      number = strtol (p->arg, (char **) &p->arg, 10);

      write_exp_elt_opcode (OP_LONG);
      write_exp_elt_type (builtin_type (gdbarch)->builtin_long);
      write_exp_elt_longcst (number);
      write_exp_elt_opcode (OP_LONG);

      if (const_sufix)
	{
	  if (strncmp (p->arg, const_sufix, const_sufix_len) == 0)
	    p->arg += const_sufix_len;
	  else
	    error (_("Invalid constant sufix on expression `%s'."),
		   p->saved_arg);
	}
    }
  else if ((reg_prefix
	    && strncmp (p->arg, reg_prefix, reg_prefix_len) == 0)
	   || (reg_ind_prefix
	       && strncmp (p->arg, reg_ind_prefix, reg_ind_prefix_len) == 0))
    stap_parse_register_operand (p);
  else
    error (_("Operator `%c' not recognized on expression `%s'."),
	   *p->arg, p->saved_arg);
}

static void
stap_parse_argument_conditionally (struct stap_parse_info *p)
{
  if (*p->arg == '-' || *p->arg == '~' || *p->arg == '+' /* Unary.  */
      || isdigit (*p->arg)
      || gdbarch_stap_is_single_operand (p->gdbarch, p->arg))
    stap_parse_single_operand (p);
  else if (*p->arg == '(')
    {
      /* We are dealing with a parenthesized operand.  It means we
	 have to parse it as it was a separate expression, without
	 left-side or precedence.  */
      ++p->arg;
      p->arg = skip_spaces_const (p->arg);
      ++p->inside_paren_p;

      stap_parse_argument_1 (p, 0, STAP_OPERAND_PREC_NONE);

      --p->inside_paren_p;
      if (*p->arg != ')')
	error (_("Missign close-paren on expression `%s'."),
	       p->saved_arg);

      ++p->arg;
      if (p->inside_paren_p)
	p->arg = skip_spaces_const (p->arg);
    }
  else
    error (_("Cannot parse expression `%s'."), p->saved_arg);
}

static void
stap_parse_argument_1 (struct stap_parse_info *p, int has_lhs,
		       enum stap_operand_prec prec)
{
  /* This is an operator-precedence parser.

     We work with left- and right-sides of expressions, and
     parse them depending on the precedence of the operators
     we find.  */

  if (p->inside_paren_p)
    p->arg = skip_spaces_const (p->arg);

  if (!has_lhs)
    /* We were called without a left-side, either because this is the
       first call, or because we were called to parse a parenthesized
       expression.  It doesn't really matter; we have to parse the
       left-side in order to continue the process.  */
    stap_parse_argument_conditionally (p);

  /* Start to parse the right-side, and to "join" left and right sides
     depending on the operation specified.

     This loop shall continue until we run out of characters in the input,
     or until we find a close-parenthesis, which means that we've reached
     the end of a sub-expression.  */
  while (p->arg && *p->arg && *p->arg != ')' && !isspace (*p->arg))
    {
      const char *tmp_exp_buf;
      enum exp_opcode opcode;
      enum stap_operand_prec cur_prec;

      if (!stap_is_operator (*p->arg))
	error (_("Invalid operator `%c' on expression `%s'."), *p->arg,
	       p->saved_arg);

      /* We have to save the current value of the expression buffer because
	 the `stap_get_opcode' modifies it in order to get the current
	 operator.  If this operator's precedence is lower than PREC, we
	 should return and not advance the expression buffer pointer.  */
      tmp_exp_buf = p->arg;
      stap_get_opcode (&tmp_exp_buf, &opcode);

      cur_prec = stap_get_operator_prec (opcode);
      if (cur_prec < prec)
	/* If the precedence of the operator that we are seeing now is
	   lower than the precedence of the first operator seen before
	   this parsing process began, it means we should stop parsing
	   and return.  */
	break;

      p->arg = tmp_exp_buf;
      if (p->inside_paren_p)
	p->arg = skip_spaces_const (p->arg);

      /* Parse the right-side of the expression.  */
      stap_parse_argument_conditionally (p);

      /* While we still have operators, try to parse another
	 right-side, but using the current right-side as a left-side.  */
      while (*p->arg && stap_is_operator (*p->arg))
	{
	  enum exp_opcode lookahead_opcode;
	  enum stap_operand_prec lookahead_prec;

	  /* Saving the current expression buffer position.  The explanation
	     is the same as above.  */
	  tmp_exp_buf = p->arg;
	  stap_get_opcode (&tmp_exp_buf, &lookahead_opcode);
	  lookahead_prec = stap_get_operator_prec (lookahead_opcode);

	  if (lookahead_prec <= prec)
	    /* If we are dealing with an operator whose precedence is lower
	       than the first one, just abandon the attempt.  */
	    break;

	  /* Parse the right-side of the expression, but since we already
	     have a left-side at this point, set `has_lhs' to 1.  */
	  stap_parse_argument_1 (p, 1, lookahead_prec);
	}

      write_exp_elt_opcode (opcode);
    }
}

/* Parse a probe's argument.

   Assuming that:

   LP = literal integer prefix
   LS = literal integer sufix

   RP = register prefix
   RS = register sufix

   RIP = register indirection prefix
   RIS = register indirection sufix

   This routine assumes that arguments' tokens are of the form:

   - [LP] NUMBER [LS]
   - [RP] REGISTER [RS]
   - [RIP] [RP] REGISTER [RS] [RIS]
   - If we find a number without LP, we try to parse it as a literal integer
   constant (if LP == NULL), or as a register displacement.
   - We count parenthesis, and only skip whitespaces if we are inside them.
   - If we find an operator, we skip it.

   This function can also call a special function that will try to match
   unknown tokens.  It will return 1 if the argument has been parsed
   successfully, or zero otherwise.  */

static int
stap_parse_argument (const char **arg, struct type *atype,
		     struct gdbarch *gdbarch)
{
  struct stap_parse_info p;
  volatile struct gdb_exception e;

  /* We need to initialize the expression buffer, in order to begin
     our parsing efforts.  The language here does not matter, since we
     are using our own parser.  */
  initialize_expout (10, current_language, gdbarch);

  p.saved_arg = *arg;
  p.arg = *arg;
  p.arg_type = atype;
  p.gdbarch = gdbarch;
  p.inside_paren_p = 0;

  TRY_CATCH (e, RETURN_MASK_ERROR)
    {
      stap_parse_argument_1 (&p, 0, STAP_OPERAND_PREC_NONE);
    }
  if (e.reason < 0)
    {
      xfree (expout);
      return 0;
    }

  gdb_assert (p.inside_paren_p == 0);

  /* Casting the final expression to the appropriate type.  */
  write_exp_elt_opcode (UNOP_CAST);
  write_exp_elt_type (atype);
  write_exp_elt_opcode (UNOP_CAST);

  reallocate_expout ();

  p.arg = skip_spaces_const (p.arg);
  *arg = p.arg;

  return 1;
}

/* Helper function which is responsible for freeing the space allocated to
   hold information about a probe's arguments.  */

static void
stap_free_args_info (void *args_info_ptr)
{
  struct stap_args_info *a = (struct stap_args_info *) args_info_ptr;
  int i;

  for (i = 0; i < a->n_args; i++)
    xfree (a->args[i].aexpr);

  xfree (a->args);
  xfree (a);
}

/* Function which parses an argument string from PROBE, correctly splitting
   the arguments and storing their information in properly ways.

   Consider the following argument string (x86 syntax):

   `4@%eax 4@$10'

   We have two arguments, `%eax' and `$10', both with 32-bit unsigned bitness.
   This function basically handles them, properly filling some structures with
   this information.  */

static void
stap_parse_probe_arguments (struct stap_probe *probe)
{
  struct stap_args_info *args_info;
  struct cleanup *back_to;
  const char *cur = probe->args;
  int current_arg = -1;
  /* This is a state-machine parser, which means we will always be
     in a known state when parsing an argument.  The state could be
     either `NEW_ARG' if we are parsing a new argument, `BITNESS' if
     we are parsing the bitness-definition part (i.e., `4@'), or
     `PARSE_ARG' if we are actually parsing the argument part.  */
  enum
    {
      NEW_ARG,
      BITNESS,
      PARSE_ARG,
    } current_state;

  /* For now, we assume everything is not going to work.  */
  probe->parsed_args = &dummy_stap_args_info;

  if (!cur || !*cur || *cur == ':')
    return;

  args_info = xmalloc (sizeof (struct stap_args_info));
  args_info->n_args = 0;
  back_to = make_cleanup (stap_free_args_info, args_info);
  args_info->args = xcalloc (STAP_MAX_ARGS, sizeof (struct stap_probe_arg));

  /* Ok, let's start.  */
  current_state = NEW_ARG;

  while (*cur)
    {
      switch (current_state)
	{
	case NEW_ARG:
	  ++current_arg;

	  if (current_arg >= STAP_MAX_ARGS)
	    {
	      complaint (&symfile_complaints,
			 _("probe `%s' has more arguments than the maximum "
			   "allowed"), probe->name);
	      do_cleanups (back_to);
	      return;
	    }

	  current_state = BITNESS;
	  break;

	case BITNESS:
	    {
	      enum stap_arg_bitness b;
	      int got_minus = 0;

	      /* We expect to find something like:

		 N@OP

		 Where `N' can be [+,-][4,8].  This is not mandatory, so
		 we check it here.  If we don't find it, go to the next
		 state.  */
	      if ((*cur == '-' && cur[1] && cur[2] != '@')
		  && cur[1] != '@')
		{
		  current_state = PARSE_ARG;
		  args_info->args[current_arg].bitness
		    = STAP_ARG_BITNESS_UNDEFINED;
		  break;
		}

	      if (*cur == '-')
		{
		  /* Discard the `-'.  */
		  ++cur;
		  got_minus = 1;
		}

	      if (*cur == '4')
		b = got_minus ? STAP_ARG_BITNESS_32BIT_SIGNED
		  : STAP_ARG_BITNESS_32BIT_UNSIGNED;
	      else if (*cur == '8')
		b = got_minus ? STAP_ARG_BITNESS_64BIT_SIGNED
		  : STAP_ARG_BITNESS_64BIT_UNSIGNED;
	      else
		{
		  /* We have an error, because we don't expect anything
		     except 4 and 8.  */
		  complaint (&symfile_complaints,
			     _("unrecognized bitness `%c' for probe `%s'"),
			     *cur, probe->name);
		  do_cleanups (back_to);
		  return;
		}

	      args_info->args[current_arg].bitness = b;
	      args_info->args[current_arg].atype
		= stap_get_expected_argument_type (probe->gdbarch, b);
	      /* Discard the number and the `@' sign.  */
	      cur += 2;
	      /* Move on.  */
	      current_state = PARSE_ARG;
	    }
	  break;

	case PARSE_ARG:
	    {
	      if (!stap_parse_argument (&cur,
					args_info->args[current_arg].atype,
					probe->gdbarch))
		{
		  /* We have tried to parse this argument, but it's
		     malformed.  This is an error.  */
		  complaint (&symfile_complaints,
			     _("malformed argument for probe `%s'"),
			     probe->name);
		  do_cleanups (back_to);
		  return;
		}

	      if (stap_expression_debug)
		dump_raw_expression (expout, gdb_stdlog,
				     "before conversion to prefix form");

	      prefixify_expression (expout);

	      if (stap_expression_debug)
		dump_prefix_expression (expout, gdb_stdlog);

	      args_info->args[current_arg].aexpr = expout;
	      expout = NULL;

	      ++args_info->n_args;
	      /* Start it over again.  */
	      cur = skip_spaces_const (cur);
	      current_state = NEW_ARG;
	    }
	  break;
	}

      if (!*cur && current_state != NEW_ARG)
	{
	  /* We reached the end of the argument string, but we're
	     still in the middle of the process of parsing an argument.
	     It means the argument string is malformed.  */
	  complaint (&symfile_complaints,
		     _("malformed argument for probe `%s'"),
		     probe->name);
	  do_cleanups (back_to);
	  return;
	}
    }

  args_info->args = xrealloc (args_info->args,
			      args_info->n_args
			      * sizeof (struct stap_probe_arg));
  args_info->probe = probe;

  probe->parsed_args = args_info;

  discard_cleanups (back_to);
}

/* See definition in stap-probe.h.  */

int
stap_get_probe_argument_count (struct stap_probe *probe)
{
  if (!probe->parsed_args)
    stap_parse_probe_arguments (probe);

  return probe->parsed_args->n_args;
}

/* Return 1 if OP is a valid operator inside a probe argument, or zero
   otherwise.  */

static int
stap_is_operator (char op)
{
  return (op == '+' || op == '-' || op == '*' || op == '/'
	  || op == '>' || op == '<' || op == '!' || op == '^'
	  || op == '|' || op == '&' || op == '%' || op == '=');
}

/* See definition in stap-probe.h.  */

struct value *
stap_evaluate_probe_argument (struct objfile *objfile,
			      struct stap_probe *probe,
			      struct frame_info *frame,
			      int n)
{
  int pos = 0;

  if (!probe->parsed_args)
    stap_parse_probe_arguments (probe);

  if (!probe->parsed_args->args
      || n >= probe->parsed_args->n_args)
    return NULL;

  /* This is needed because on some architectures (e.g., ARM) we need
     the frame's gdbarch in order to compute the value of the frame
     pointer.  */
  probe->parsed_args->args[n].aexpr->gdbarch = get_frame_arch (frame);

  return evaluate_subexp_standard (probe->parsed_args->args[n].atype,
				   probe->parsed_args->args[n].aexpr,
				   &pos, EVAL_NORMAL);
}

/* See definition in stap-probe.h.  */

void
stap_compile_to_ax (struct objfile *objfile,
		    struct stap_probe *probe,
		    struct agent_expr *expr,
		    struct axs_value *value,
		    int n)
{
  union exp_element *pc;

  if (!probe->parsed_args)
    stap_parse_probe_arguments (probe);

  if (!probe->parsed_args->args
      || n >= probe->parsed_args->n_args)
    return;

  pc = probe->parsed_args->args[n].aexpr->elts;
  gen_expr (probe->parsed_args->args[n].aexpr, &pc, expr, value);

  require_rvalue (expr, value);
  value->type = probe->parsed_args->args[n].atype;
}

struct value *
stap_safe_evaluate_at_pc (struct frame_info *frame, int n)
{
  struct stap_probe *probe;
  struct objfile *objfile;
  int n_probes;

  probe = find_probe_by_pc (get_frame_pc (frame), &objfile);
  if (!probe)
    return NULL;
  gdb_assert (objfile->sf && objfile->sf->sym_probe_fns);

  n_probes
    = objfile->sf->sym_probe_fns->sym_get_probe_argument_count (objfile,
								probe);
  if (n >= n_probes)
    return NULL;

  return objfile->sf->sym_probe_fns->sym_evaluate_probe_argument (objfile,
								  probe,
								  frame,
								  n);
}

/* This function frees the space allocated to hold information about
   the probe's parsed arguments.  */

void
stap_free_parsed_args (struct stap_args_info *parsed_args)
{
  int i;

  if (!parsed_args
      || parsed_args == &dummy_stap_args_info
      || parsed_args->n_args == 0)
    return;

  for (i = 0; i < parsed_args->n_args; i++)
    xfree (parsed_args->args[i].aexpr);

  xfree (parsed_args->args);
  xfree (parsed_args);
}

/* A utility structure.  A VEC of these is built when handling "info
   probes".  */

struct stap_probe_and_objfile
{
  /* The probe.  */
  struct stap_probe *probe;
  /* The probe's objfile.  */
  struct objfile *objfile;
};

typedef struct stap_probe_and_objfile stap_entry;
DEF_VEC_O (stap_entry);

/* A helper function for collect_probes that compiles a regexp and
   throws an exception on error.  This installs a cleanup to free the
   resulting pattern on success.  If RX is NULL, this does nothing.  */

static void
compile_rx_or_error (regex_t *pattern, const char *rx, const char *message)
{
  int code;

  if (!rx)
    return;

  code = regcomp (pattern, rx, REG_NOSUB);
  if (code == 0)
    make_regfree_cleanup (pattern);
  else
    {
      char *err = get_regcomp_error (code, pattern);

      make_cleanup (xfree, err);
      error (_("%s: %s"), message, err);
    }
}

/* Make a vector of probes matching OBJNAME, PROVIDER, and PROBE.
   Each argument is a regexp, or NULL, which matches anything.  */

static VEC (stap_entry) *
collect_probes (char *objname, char *provider, char *probe)
{
  struct objfile *objfile;
  VEC (stap_entry) *result = NULL;
  struct cleanup *cleanup;
  regex_t obj_pat, prov_pat, probe_pat;

  cleanup = make_cleanup (VEC_cleanup (stap_entry), &result);

  compile_rx_or_error (&prov_pat, provider, _("Invalid provider regexp"));
  compile_rx_or_error (&probe_pat, probe, _("Invalid probe regexp"));
  compile_rx_or_error (&obj_pat, objname, _("Invalid object file regexp"));

  ALL_OBJFILES (objfile)
  {
    struct stap_probe *probes;
    int i, num_probes;

    if (! objfile->sf || ! objfile->sf->sym_probe_fns)
      continue;

    if (objname)
      {
	if (regexec (&obj_pat, objfile->name, 0, NULL, 0) != 0)
	  continue;
      }

    probes = objfile->sf->sym_probe_fns->sym_get_probes (objfile, &num_probes);
    for (i = 0; i < num_probes; ++i)
      {
	stap_entry entry;

	if (provider)
	  {
	    if (regexec (&prov_pat, probes[i].provider, 0, NULL, 0) != 0)
	      continue;
	  }

	if (probe)
	  {
	    if (regexec (&probe_pat, probes[i].name, 0, NULL, 0) != 0)
	      continue;
	  }

	entry.probe = &probes[i];
	entry.objfile = objfile;
	VEC_safe_push (stap_entry, result, &entry);
      }
  }

  discard_cleanups (cleanup);
  return result;
}

/* A qsort comparison function for stap_entry objects.  */

static int
compare_entries (const void *a, const void *b)
{
  const stap_entry *ea = a;
  const stap_entry *eb = b;
  int v;

  v = strcmp (ea->probe->provider, eb->probe->provider);
  if (v)
    return v;

  v = strcmp (ea->probe->name, eb->probe->name);
  if (v)
    return v;

  if (ea->probe->address < eb->probe->address)
    return -1;
  if (ea->probe->address > eb->probe->address)
    return 1;

  return strcmp (ea->objfile->name, eb->objfile->name);
}

/* Implementation of the "info probes" command.  */

static void
info_probes_command (char *arg, int from_tty)
{
  char *provider, *probe = NULL, *objname = NULL;
  struct cleanup *cleanup = make_cleanup (null_cleanup, NULL);
  VEC (stap_entry) *items;
  int i, addr_width, any_found;
  stap_entry *entry;

  provider = extract_arg (&arg);
  if (provider)
    {
      make_cleanup (xfree, provider);

      probe = extract_arg (&arg);
      if (probe)
	{
	  make_cleanup (xfree, probe);

	  objname = extract_arg (&arg);
	  if (objname)
	    make_cleanup (xfree, objname);
	}
    }

  items = collect_probes (objname, provider, probe);
  make_cleanup (VEC_cleanup (stap_entry), &items);
  make_cleanup_ui_out_table_begin_end (current_uiout, 5,
				       VEC_length (stap_entry, items),
				       "SystemTapProbes");

  if (! VEC_empty (stap_entry, items))
    qsort (VEC_address (stap_entry, items),
	   VEC_length (stap_entry, items),
	   sizeof (stap_entry),
	   compare_entries);

  addr_width = 4 + (gdbarch_ptr_bit (get_current_arch ()) / 4);

  ui_out_table_header (current_uiout, 10, ui_left, "provider", _("Provider"));
  ui_out_table_header (current_uiout, 10, ui_left, "name", _("Name"));
  ui_out_table_header (current_uiout, addr_width - 1, ui_left, "addr", _("Where"));
  ui_out_table_header (current_uiout, addr_width - 1, ui_left, "semaphore",
		       _("Semaphore"));
  ui_out_table_header (current_uiout, 30, ui_left, "object", _("Object"));
  ui_out_table_body (current_uiout);

  for (i = 0; VEC_iterate (stap_entry, items, i, entry); ++i)
    {
      struct cleanup *inner;

      inner = make_cleanup_ui_out_tuple_begin_end (current_uiout, "probe");

      ui_out_field_string (current_uiout, "provider", entry->probe->provider);
      ui_out_field_string (current_uiout, "name", entry->probe->name);
      ui_out_field_core_addr (current_uiout, "addr", get_current_arch (),
			      entry->probe->address);
      if (entry->probe->sem_addr == 0)
	ui_out_field_skip (current_uiout, "semaphore");
      else
      ui_out_field_core_addr (current_uiout, "semaphore", get_current_arch (),
			      entry->probe->sem_addr);
      ui_out_field_string (current_uiout, "object", entry->objfile->name);
      ui_out_text (current_uiout, "\n");

      do_cleanups (inner);
    }

  any_found = ! VEC_empty (stap_entry, items);
  do_cleanups (cleanup);

  if (! any_found)
    ui_out_message (current_uiout, 0, _("No probes matched.\n"));
}



/* See definition in stap-probe.h.  */

VEC (stap_probe_p) *
find_probes_in_objfile (struct objfile *objfile,
			const char *provider,
			const char *name)
{
  struct stap_probe *probes;
  int i, num_probes;
  VEC (stap_probe_p) *result = NULL;

  if (! objfile->sf || ! objfile->sf->sym_probe_fns)
    return NULL;

  probes = objfile->sf->sym_probe_fns->sym_get_probes (objfile, &num_probes);
  for (i = 0; i < num_probes; ++i)
    {
      if (strcmp (probes[i].provider, provider) != 0)
	continue;

      if (strcmp (probes[i].name, name) != 0)
	continue;

      VEC_safe_push (stap_probe_p, result, &probes[i]);
    }

  return result;
}

/* See definition in stap-probe.h.  */

struct symtabs_and_lines
parse_stap_probe (char **argptr, struct linespec_result *canonical)
{
  char *arg_start, *arg_end, *arg;
  char *objfile_name = NULL, *provider = NULL, *name, *p;
  struct cleanup *cleanup;
  struct symtabs_and_lines result;
  struct objfile *objfile;

  result.sals = NULL;
  result.nelts = 0;

  arg_start = *argptr;
  /* The caller ensured that this starts with '-p'.  */
  gdb_assert (arg_start && strncmp (arg_start, "-p", 2) == 0);
  arg_end = arg_start + 2;
  arg_end = skip_spaces (arg_end);

  if (!*arg_end)
    error (_("argument to `-p' missing"));

  arg = arg_end;
  arg_end = skip_to_space (arg_end);

  /* We make a copy here so we can write over parts with impunity.  */
  arg = savestring (arg, arg_end - arg);
  cleanup = make_cleanup (xfree, arg);

  /* Extract each word from the argument, separated by ":"s.  */
  p = strchr (arg, ':');
  if (p == NULL)
    {
      /* This is `-p name'.  */
      name = arg;
    }
  else
    {
      char *hold = p + 1;

      *p = '\0';
      p = strchr (hold, ':');
      if (p == NULL)
	{
	  /* This is `-p provider:name'.  */
	  provider = arg;
	  name = hold;
	}
      else
	{
	  /* This is `-p objfile:provider:name'.  */
	  *p = '\0';
	  objfile_name = arg;
	  provider = hold;
	  name = p + 1;
	}
    }

  if (*name == '\0')
    error (_("no probe name specified"));
  if (provider && *provider == '\0')
    error (_("invalid provider name"));
  if (objfile_name && *objfile_name == '\0')
    error (_("invalid objfile name"));

  ALL_OBJFILES (objfile)
  {
    struct stap_probe *probes;
    int i, num_probes;

    if (! objfile->sf || ! objfile->sf->sym_probe_fns)
      continue;

    if (objfile_name
	&& FILENAME_CMP (objfile->name, objfile_name) != 0
	&& FILENAME_CMP (lbasename (objfile->name), objfile_name) != 0)
      continue;

    probes = objfile->sf->sym_probe_fns->sym_get_probes (objfile, &num_probes);
    for (i = 0; i < num_probes; ++i)
      {
	struct symtab_and_line *sal;

	if (provider && strcmp (probes[i].provider, provider) != 0)
	  continue;

	if (strcmp (probes[i].name, name) != 0)
	  continue;

	++result.nelts;
	result.sals = xrealloc (result.sals,
				result.nelts * sizeof (struct symtab_and_line));
	sal = &result.sals[result.nelts - 1];

	init_sal (sal);

	sal->pc = probes[i].address;
	sal->explicit_pc = 1;
	sal->section = find_pc_overlay (sal->pc);
	sal->pspace = current_program_space;
	sal->semaphore = probes[i].sem_addr;
      }
  }

  if (result.nelts == 0)
    {
      throw_error (NOT_FOUND_ERROR,
		   _("No probe matching objfile=`%s', provider=`%s', name=`%s'"),
		   objfile_name ? objfile_name : _("<any>"),
		   provider ? provider : _("<any>"),
		   name);
    }

  if (canonical)
    {
      canonical->special_display = 1;
      canonical->pre_expanded = 1;
      canonical->addr_string = savestring (*argptr, arg_end - *argptr);
    }

  *argptr = arg_end;
  do_cleanups (cleanup);

  return result;
}



/* See definition in stap-probe.h.  */

struct stap_probe *
find_probe_by_pc (CORE_ADDR pc, struct objfile **objfile_out)
{
  struct objfile *objfile;

  ALL_OBJFILES (objfile)
  {
    struct stap_probe *probes;
    int i, num_probes;
    stap_entry entry;

    if (! objfile->sf || ! objfile->sf->sym_probe_fns)
      continue;

    /* If this proves too inefficient, we can replace with a hash.  */
    probes = objfile->sf->sym_probe_fns->sym_get_probes (objfile, &num_probes);
    for (i = 0; i < num_probes; ++i)
      {
	if (probes[i].address == pc)
	  {
	    *objfile_out = objfile;
	    return &probes[i];
	  }
      }
  }

  return NULL;
}

/* This is called to compute the value of one of the $_probe_arg*
   convenience variables.  */

static struct value *
compute_probe_arg (struct gdbarch *arch, struct internalvar *ivar,
		   void *data)
{
  struct frame_info *frame = get_selected_frame (_("No frame selected"));
  CORE_ADDR pc = get_frame_pc (frame);
  int sel = (int) (uintptr_t) data;
  struct objfile *objfile;
  struct stap_probe *pc_probe;
  int n_probes;

  /* SEL==-1 means "_probe_argc".  */
  gdb_assert (sel >= -1 && sel <= STAP_MAX_ARGS);

  pc_probe = find_probe_by_pc (pc, &objfile);
  if (pc_probe == NULL)
    error (_("No SystemTap probe at PC %s"), core_addr_to_string (pc));

  n_probes
    = objfile->sf->sym_probe_fns->sym_get_probe_argument_count (objfile,
								pc_probe);
  if (sel == -1)
    return value_from_longest (builtin_type (arch)->builtin_int, n_probes);

  if (sel >= n_probes)
    error (_("Invalid probe argument %d -- probe has %d arguments available"),
	   sel, n_probes);

  return objfile->sf->sym_probe_fns->sym_evaluate_probe_argument (objfile,
								  pc_probe,
								  frame, sel);
}

/* This is called to compile one of the $_probe_arg* convenience
   variables into an agent expression.  */

static void
compile_probe_arg (struct internalvar *ivar, struct agent_expr *expr,
		   struct axs_value *value, void *data)
{
  CORE_ADDR pc = expr->scope;
  int sel = (int) (uintptr_t) data;
  struct objfile *objfile;
  struct stap_probe *pc_probe;
  int n_probes;

  /* SEL==-1 means "_probe_argc".  */
  gdb_assert (sel >= -1 && sel <= STAP_MAX_ARGS);

  pc_probe = find_probe_by_pc (pc, &objfile);
  if (pc_probe == NULL)
    error (_("No SystemTap probe at PC %s"), core_addr_to_string (pc));

  n_probes
    = objfile->sf->sym_probe_fns->sym_get_probe_argument_count (objfile,
								pc_probe);
  if (sel == -1)
    {
      value->kind = axs_rvalue;
      value->type = builtin_type (expr->gdbarch)->builtin_int;
      ax_const_l (expr, n_probes);
      return;
    }

  gdb_assert (sel >= 0);
  if (sel >= n_probes)
    error (_("Invalid probe argument %d -- probe has %d arguments available"),
	   sel, n_probes);

  objfile->sf->sym_probe_fns->sym_compile_to_ax (objfile, pc_probe,
						 expr, value, sel);
}



/* Implementation of `$_probe_arg*' set of variables.  */

static const struct internalvar_funcs probe_funcs =
{
  compute_probe_arg,
  compile_probe_arg,
  NULL
};

void
_initialize_stap_probe (void)
{
  add_info ("probes", info_probes_command, _("\
Show available static probes.\n\
Usage: info probes [PROVIDER [NAME [OBJECT]]]\n\
Each argument is a regular expression, used to select probes.\n\
PROVIDER matches probe provider names.\n\
NAME matches the probe names.\n\
OBJECT match the executable or shared library name."));

  add_setshow_zinteger_cmd ("stap-expression", class_maintenance,
			    &stap_expression_debug,
			    _("Set SystemTap expression debugging."),
			    _("Show SystemTap expression debugging."),
			    _("When non-zero, the internal representation "
			      "of SystemTap expressions will be printed."),
			    NULL,
			    show_stapexpressiondebug,
			    &setdebuglist, &showdebuglist);

  create_internalvar_type_lazy ("_probe_argc", &probe_funcs,
				(void *) (uintptr_t) -1);
  create_internalvar_type_lazy ("_probe_arg0", &probe_funcs,
				(void *) (uintptr_t) 0);
  create_internalvar_type_lazy ("_probe_arg1", &probe_funcs,
				(void *) (uintptr_t) 1);
  create_internalvar_type_lazy ("_probe_arg2", &probe_funcs,
				(void *) (uintptr_t) 2);
  create_internalvar_type_lazy ("_probe_arg3", &probe_funcs,
				(void *) (uintptr_t) 3);
  create_internalvar_type_lazy ("_probe_arg4", &probe_funcs,
				(void *) (uintptr_t) 4);
  create_internalvar_type_lazy ("_probe_arg5", &probe_funcs,
				(void *) (uintptr_t) 5);
  create_internalvar_type_lazy ("_probe_arg6", &probe_funcs,
				(void *) (uintptr_t) 6);
  create_internalvar_type_lazy ("_probe_arg7", &probe_funcs,
				(void *) (uintptr_t) 7);
  create_internalvar_type_lazy ("_probe_arg8", &probe_funcs,
				(void *) (uintptr_t) 8);
  create_internalvar_type_lazy ("_probe_arg9", &probe_funcs,
				(void *) (uintptr_t) 9);
  create_internalvar_type_lazy ("_probe_arg10", &probe_funcs,
				(void *) (uintptr_t) 10);
  create_internalvar_type_lazy ("_probe_arg11", &probe_funcs,
				(void *) (uintptr_t) 11);
}
